import io
import json
import pathlib

from drf_yasg.utils import swagger_auto_schema
from rest_framework.decorators import action
from rest_framework.exceptions import APIException, ValidationError
from rest_framework.request import Request
from rest_framework.response import Response

from django_large_image import tilesource
from django_large_image.rest import params
from django_large_image.rest.base import LargeImageMixinBase

try:
    import tifftools
    from tifftools.exceptions import TifftoolsError
except ImportError:  # pragma: no cover
    tifftools = None
    TifftoolsError = None

metadata_summary = 'Returns tile metadata.'
metadata_parameters = params.BASE
metadata_internal_summary = 'Returns additional known metadata about the tile source.'
metadata_internal_parameters = params.BASE
bands_summary = 'Returns bands information.'
bands_parameters = params.BASE
band_summary = 'Returns single band information.'
band_parameters = params.BASE + [params.band]
frames_summary = 'Retrieve all channels/bands for each frame. This is used to generate a UI to control how the image is displayed.'
frames_parameters = params.BASE
tiffdump_summary = 'Returns tifftools tiffdump JSON. This will raise a `ValidationError` if the image is not a Tiff.'


class MetaDataMixin(LargeImageMixinBase):
    @swagger_auto_schema(
        method='GET',
        operation_summary=metadata_summary,
        manual_parameters=metadata_parameters,
    )
    @action(detail=False)
    def metadata(self, request: Request, pk: int = None) -> Response:
        source = self.get_tile_source(request, pk, style=False)
        metadata = tilesource.get_metadata(source)
        return Response(metadata)

    @swagger_auto_schema(
        method='GET',
        operation_summary=metadata_internal_summary,
        manual_parameters=metadata_internal_parameters,
    )
    @action(detail=False)
    def metadata_internal(self, request: Request, pk: int = None) -> Response:
        source = self.get_tile_source(request, pk, style=False)
        metadata = tilesource.get_metadata_internal(source)
        return Response(metadata)

    @swagger_auto_schema(
        method='GET',
        operation_summary=bands_summary,
        manual_parameters=bands_parameters,
    )
    @action(detail=False)
    def bands(self, request: Request, pk: int = None) -> Response:
        source = self.get_tile_source(request, pk, style=False)
        metadata = source.getBandInformation()
        return Response(metadata)

    @swagger_auto_schema(
        method='GET',
        operation_summary=bands_summary,
        manual_parameters=band_parameters,
    )
    @action(detail=False)
    def band(self, request: Request, pk: int = None) -> Response:
        # TODO: handle frame choice
        band = int(self.get_query_param(request, 'band', 1))
        source = self.get_tile_source(request, pk, style=False)
        metadata = source.getOneBandInformation(band)
        return Response(metadata)

    @swagger_auto_schema(
        method='GET',
        operation_summary=frames_summary,
        manual_parameters=frames_parameters,
    )
    @action(detail=False)
    def frames(self, request: Request, pk: int = None) -> Response:
        source = self.get_tile_source(request, pk, style=False)
        data = tilesource.get_frames(source)
        return Response(data)

    @swagger_auto_schema(
        method='GET',
        operation_summary=tiffdump_summary,
    )
    @action(detail=False)
    def tiffdump(self, request: Request, pk: int = None) -> Response:
        if tifftools is None:  # pragma: no cover
            raise APIException('`tifftools` is not installed on the server.')
        source = self.get_tile_source(request, pk, style=False)
        # This will only work for local files (path on disk)
        path = source._getLargeImagePath()
        try:
            if not pathlib.Path(path).exists():
                raise OSError  # pragma: no cover
        except OSError:
            raise APIException(
                'The image path is not local and tifftools will not be able to open this image.'
            )
        output = io.StringIO()
        try:
            tifftools.tiff_dump(path, dest=output, outformat='json')
        except (TifftoolsError, OSError) as e:
            raise ValidationError(str(e))
        output.seek(0)
        return Response(json.loads(output.read()))


class MetaDataDetailMixin(MetaDataMixin):
    @swagger_auto_schema(
        method='GET',
        operation_summary=metadata_summary,
        manual_parameters=metadata_parameters,
    )
    @action(detail=True)
    def metadata(self, request: Request, pk: int = None) -> Response:
        return super().metadata(request, pk)

    @swagger_auto_schema(
        method='GET',
        operation_summary=metadata_internal_summary,
        manual_parameters=metadata_internal_parameters,
    )
    @action(detail=True)
    def metadata_internal(self, request: Request, pk: int = None) -> Response:
        return super().metadata_internal(request, pk)

    @swagger_auto_schema(
        method='GET',
        operation_summary=bands_summary,
        manual_parameters=bands_parameters,
    )
    @action(detail=True)
    def bands(self, request: Request, pk: int = None) -> Response:
        return super().bands(request, pk)

    @swagger_auto_schema(
        method='GET',
        operation_summary=bands_summary,
        manual_parameters=band_parameters,
    )
    @action(detail=True)
    def band(self, request: Request, pk: int = None) -> Response:
        return super().band(request, pk)

    @swagger_auto_schema(
        method='GET',
        operation_summary=frames_summary,
        manual_parameters=frames_parameters,
    )
    @action(detail=True)
    def frames(self, request: Request, pk: int = None) -> Response:
        return super().frames(request, pk)

    @swagger_auto_schema(
        method='GET',
        operation_summary=tiffdump_summary,
    )
    @action(detail=True)
    def tiffdump(self, request: Request, pk: int = None) -> Response:
        return super().tiffdump(request, pk)



from django.http import HttpResponse
from django.utils.decorators import method_decorator
from django.views.decorators.cache import cache_page
from drf_yasg.utils import swagger_auto_schema
from large_image.exceptions import TileSourceXYZRangeError
from rest_framework.decorators import action
from rest_framework.exceptions import ValidationError
from rest_framework.request import Request
from rest_framework.response import Response

from django_large_image import tilesource
from django_large_image.rest import params
from django_large_image.rest.base import CACHE_TIMEOUT, LargeImageMixinBase
from django_large_image.rest.serializers import TileMetadataSerializer

tile_metadata_summary = 'Returns tile metadata.'
tile_metadata_parameters = params.BASE
tile_summary = 'Returns tile image binary.'
tile_parameters = params.BASE + [params.z, params.x, params.y] + params.STYLE
tile_corners_summary = 'Returns bounds of a tile for a given x, y, z index.'
tile_corners_parameters = params.BASE + [params.z, params.x, params.y]


class TilesMixin(LargeImageMixinBase):
    @swagger_auto_schema(
        method='GET',
        operation_summary=tile_metadata_summary,
        manual_parameters=tile_metadata_parameters,
    )
    @action(detail=False, url_path=r'tiles/metadata')
    def tiles_metadata(self, request: Request, pk: int = None) -> Response:
        source = self.get_tile_source(request, pk, style=False)
        source.dli_geospatial = tilesource.is_geospatial(source)
        serializer = TileMetadataSerializer(source)
        return Response(serializer.data)

    @method_decorator(cache_page(CACHE_TIMEOUT))
    def tile(
        self, request: Request, x: int, y: int, z: int, pk: int = None, format: str = None
    ) -> HttpResponse:
        encoding = tilesource.format_to_encoding(format)
        source = self.get_tile_source(request, pk, encoding=encoding)
        try:
            tile_binary = source.getTile(int(x), int(y), int(z), encoding=encoding)
        except TileSourceXYZRangeError as e:
            raise ValidationError(e)
        mime_type = source.getTileMimeType()
        return HttpResponse(tile_binary, content_type=mime_type)

    @swagger_auto_schema(
        method='GET',
        operation_summary=tile_summary,
        manual_parameters=tile_parameters,
    )
    @action(detail=False, url_path=r'tiles/(?P<z>\d+)/(?P<x>\d+)/(?P<y>\d+).png')
    def tile_png(
        self,
        request: Request,
        x: int,
        y: int,
        z: int,
        pk: int = None,
    ) -> HttpResponse:
        return self.tile(request, x, y, z, pk=pk, format='png')

    @swagger_auto_schema(
        method='GET',
        operation_summary=tile_summary,
        manual_parameters=tile_parameters,
    )
    @action(detail=False, url_path=r'tiles/(?P<z>\d+)/(?P<x>\d+)/(?P<y>\d+).jpeg')
    def tile_jpeg(
        self,
        request: Request,
        x: int,
        y: int,
        z: int,
        pk: int = None,
    ) -> HttpResponse:
        return self.tile(request, x, y, z, pk=pk, format='jpeg')

    @swagger_auto_schema(
        method='GET',
        operation_summary=tile_corners_summary,
        manual_parameters=tile_corners_parameters,
    )
    @action(
        detail=False, methods=['get'], url_path=r'tiles/(?P<z>\d+)/(?P<x>\d+)/(?P<y>\d+)/corners'
    )
    def tile_corners(
        self, request: Request, x: int, y: int, z: int, pk: int = None
    ) -> HttpResponse:
        source = self.get_tile_source(request, pk, style=False)
        xmin, ymin, xmax, ymax = source.getTileCorners(int(z), int(x), int(y))
        metadata = {
            'xmin': xmin,
            'xmax': xmax,
            'ymin': ymin,
            'ymax': ymax,
            'proj4': source.getProj4String(),
        }
        return Response(metadata)


class TilesDetailMixin(TilesMixin):
    @swagger_auto_schema(
        method='GET',
        operation_summary=tile_metadata_summary,
        manual_parameters=tile_metadata_parameters,
    )
    @action(detail=True, url_path=r'tiles/metadata')
    def tiles_metadata(self, request: Request, pk: int = None) -> Response:
        return super().tiles_metadata(request, pk)

    @swagger_auto_schema(
        method='GET',
        operation_summary=tile_summary,
        manual_parameters=tile_parameters,
    )
    @action(detail=True, url_path=r'tiles/(?P<z>\d+)/(?P<x>\d+)/(?P<y>\d+).png')
    def tile_png(
        self,
        request: Request,
        x: int,
        y: int,
        z: int,
        pk: int = None,
    ) -> HttpResponse:
        return super().tile_png(request, x, y, z, pk)

    @swagger_auto_schema(
        method='GET',
        operation_summary=tile_summary,
        manual_parameters=tile_parameters,
    )
    @action(detail=True, url_path=r'tiles/(?P<z>\d+)/(?P<x>\d+)/(?P<y>\d+).jpeg')
    def tile_jpeg(
        self,
        request: Request,
        x: int,
        y: int,
        z: int,
        pk: int = None,
    ) -> HttpResponse:
        return super().tile_jpeg(request, x, y, z, pk)

    @swagger_auto_schema(
        method='GET',
        operation_summary=tile_corners_summary,
        manual_parameters=tile_corners_parameters,
    )
    @action(
        detail=True, methods=['get'], url_path=r'tiles/(?P<z>\d+)/(?P<x>\d+)/(?P<y>\d+)/corners'
    )
    def tile_corners(
        self, request: Request, x: int, y: int, z: int, pk: int = None
    ) -> HttpResponse:
        return super().tile_corners(request, x, y, z, pk)



from contextlib import contextmanager
import logging
import os
import pathlib
import shutil
import tempfile
from urllib.parse import urlencode

from django.conf import settings
from django.core.files import File
from django.db.models.fields.files import FieldFile
from filelock import FileLock

try:
    from minio_storage.storage import MinioStorage  # pragma: no cover
except ImportError:  # pragma: no cover
    MinioStorage = None

logger = logging.getLogger(__name__)


def param_nully(value) -> bool:
    """Determine null-like values."""
    if isinstance(value, str):
        value = value.lower()
    return value in [None, '', 'undefined', 'none', 'null', 'false']


@contextmanager
def patch_internal_presign(f: FieldFile):
    """Create an environment where Minio-based `FieldFile`s construct a locally accessible presigned URL.

    Sometimes the external host differs from the internal host for Minio files (e.g. in development).
    Getting the URL in this context ensures that the presigned URL returns the correct host for the
    odd situation of accessing the file locally.

    Note
    ----
    If concerned regarding concurrent access, see https://github.com/ResonantGeoData/ResonantGeoData/issues/287

    """
    if (
        MinioStorage is not None
        and isinstance(f.storage, MinioStorage)
        and getattr(settings, 'MINIO_STORAGE_MEDIA_URL', None) is not None
    ):
        original_base_url = f.storage.base_url
        if f.storage.base_url is not None:
            try:
                f.storage.base_url = None
                yield
            finally:
                f.storage.base_url = original_base_url
            return
    yield


def get_temp_dir() -> pathlib.Path:
    path = pathlib.Path(
        getattr(
            settings, 'DATA_TEMP_DIR', os.path.join(tempfile.gettempdir(), 'django-large-image')
        )
    )
    path.mkdir(parents=True, exist_ok=True)
    return path


def get_cache_dir() -> pathlib.Path:
    path = pathlib.Path(get_temp_dir(), 'file_cache')
    path.mkdir(parents=True, exist_ok=True)
    return path


def make_vsi(url: str, **options) -> str:
    if str(url).startswith('s3://'):
        s3_path = url.replace('s3://', '')
        vsi = f'/vsis3/{s3_path}'
    else:
        gdal_options = {
            'url': str(url),
            'use_head': 'no',
            'list_dir': 'no',
        }
        gdal_options.update(options)
        vsi = f'/vsicurl?{urlencode(gdal_options)}'
    return vsi


def get_lock_dir() -> pathlib.Path:
    path = pathlib.Path(get_temp_dir(), 'file_locks')
    path.mkdir(parents=True, exist_ok=True)
    return path


def get_file_lock(path: pathlib.Path) -> FileLock:
    """Create a file lock under the lock directory."""
    # Computes the hash using Pathlib's hash implementation on absolute path
    sha = hash(path.absolute())
    lock_path = pathlib.Path(get_lock_dir(), f'{sha}.lock')
    lock = FileLock(str(lock_path))
    return lock


def get_file_safe_path(path: pathlib.Path) -> pathlib.Path:
    """Mark the file/lock as safe to use."""
    sha = hash(path.absolute())
    return pathlib.Path(get_lock_dir(), f'{sha}.safe')


def field_file_to_local_path(field_file: FieldFile) -> pathlib.Path:
    """Download entire FieldFile to disk location.

    This overrides `girder_utils.field_file_to_local_path` to download file to
    local path without a context manager. Cleanup must be handled by caller.

    This puts a file lock on the file to prevent concurrent access while
    downloading.

    """
    field_file_basename = pathlib.PurePath(field_file.name).name
    directory = get_cache_dir() / f'{type(field_file.instance).__name__}-{field_file.instance.pk}'
    dest_path = directory / field_file_basename

    lock = get_file_lock(dest_path)
    safe = get_file_safe_path(dest_path)

    with lock.acquire():
        if not safe.exists():
            # file doesn't yet exist (or is corrupt) and we need to download it
            dest_path.parent.mkdir(parents=True, exist_ok=True)
            with field_file.open('rb'):
                file_obj: File = field_file.file
                if type(file_obj) is File:
                    # When file_obj is an actual File, (typically backed by FileSystemStorage),
                    # it is already at a stable path on disk.
                    # We must symlink it into the desired path
                    os.symlink(file_obj.name, dest_path)
                    logger.debug('Performing symlink')
                else:
                    # When file_obj is actually a subclass of File, it only provides a Python
                    # file-like object API. So, it must be copied to a stable path.
                    logger.debug('copying file...')
                    with open(dest_path, 'wb') as dest_stream:
                        shutil.copyfileobj(file_obj, dest_stream)
                        dest_stream.flush()
            # Mark file as safe
            logger.debug('Marking as safely downloaded...')
            safe.touch()
    return dest_path



import logging

import large_image
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.views import APIView

logger = logging.getLogger(__name__)


class ListTileSourcesView(APIView):
    def get(self, request: Request) -> Response:
        large_image.tilesource.loadTileSources()
        sources = large_image.tilesource.AvailableTileSources
        return Response({k: str(v) for k, v in sources.items()})


class ListColormapsView(APIView):
    def get(self, request: Request) -> Response:
        """List of available palettes.

        This does not currently list the palettable palettes there isn't a clean
        way to list all of them.
        """
        simple = {
            'red': ['#000', '#f00'],
            'r': ['#000', '#f00'],
            'green': ['#000', '#0f0'],
            'g': ['#000', '#0f0'],
            'blue': ['#000', '#00f'],
            'b': ['#000', '#00f'],
        }
        cmaps = {}
        try:
            import matplotlib.pyplot

            cmaps['matplotlib'] = list(matplotlib.pyplot.colormaps())
        except ImportError:  # pragma: no cover
            logger.error('Install matplotlib for additional colormap choices.')
        cmaps['simple'] = [s for s in simple.keys() if len(s) > 1]
        return Response(cmaps)



"""large-image utilities."""
import os
import pathlib
import tempfile
from typing import List, Optional, Tuple, Union

import large_image
from large_image.tilesource import FileTileSource
from rest_framework.exceptions import ValidationError

from django_large_image import utilities


def get_tilesource_from_path(
    path: str,
    projection: Optional[str] = None,
    style: Optional[str] = None,
    encoding: Optional[str] = None,
    source: Optional[str] = None,
) -> FileTileSource:
    if not encoding:
        encoding = 'PNG'
    if source:
        large_image.tilesource.loadTileSources()
        sources = large_image.tilesource.AvailableTileSources
        try:
            reader = sources[source]
        except KeyError:
            raise ValidationError(
                f'{source!r} is not a valid source. Try one of: {list(sources.keys())}'
            )
    else:
        reader = large_image.open
    return reader(str(path), projection=projection, style=style, encoding=encoding)


def is_geospatial(source: FileTileSource) -> bool:
    return source.getMetadata().get('geospatial', False)


def get_bounds(
    source: FileTileSource,
    projection: str = 'EPSG:4326',
) -> Optional[List[float]]:
    bounds = source.getBounds(srs=projection)
    if not bounds:
        return None
    threshold = 89.9999
    for key in ('ymin', 'ymax'):
        bounds[key] = max(min(bounds[key], threshold), -threshold)
    return bounds


def _metadata_helper(source: FileTileSource, metadata: dict):
    metadata.setdefault('geospatial', is_geospatial(source))
    if metadata['geospatial']:
        metadata['bounds'] = get_bounds(source)
        metadata['proj4'] = (source.getProj4String(),)
    if 'frames' not in metadata:
        metadata['frames'] = False


def get_metadata(source: FileTileSource) -> dict:
    metadata = source.getMetadata()
    _metadata_helper(source, metadata)
    return metadata


def get_metadata_internal(source: FileTileSource) -> dict:
    metadata = source.getInternalMetadata()
    _metadata_helper(source, metadata)
    return metadata


def _get_region(source: FileTileSource, region: dict, encoding: str) -> Tuple[pathlib.Path, str]:
    result, mime_type = source.getRegion(region=region, encoding=encoding)
    if encoding == 'TILED':
        path = result
    else:
        # Write content to temporary file
        fd, path = tempfile.mkstemp(
            suffix=f'.{encoding}', prefix='pixelRegion_', dir=str(utilities.get_cache_dir())
        )
        os.close(fd)
        path = pathlib.Path(path)
        with open(path, 'wb') as f:
            f.write(result)
    return path, mime_type


def get_region(
    source: FileTileSource,
    left: Union[float, int],
    right: Union[float, int],
    bottom: Union[float, int],
    top: Union[float, int],
    units: str = None,
    encoding: str = None,
) -> Tuple[pathlib.Path, str]:
    if isinstance(units, str):
        units = units.lower()
    if not encoding and is_geospatial(source):
        # Use tiled encoding by default for geospatial rasters
        #   output will be a tiled TIF
        encoding = 'TILED'
    elif not encoding:
        # Use JPEG encoding by default for nongeospatial rasters
        encoding = 'JPEG'
    if is_geospatial(source) and units not in [
        'pixels',
        'pixel',
    ]:
        if not units:
            units = 'EPSG:4326'
        region = dict(left=left, right=right, bottom=bottom, top=top, units=units)
        return _get_region(source, region, encoding)
    units = 'pixels'
    left, right = min(left, right), max(left, right)
    top, bottom = min(top, bottom), max(top, bottom)
    region = dict(left=left, right=right, bottom=bottom, top=top, units=units)
    return _get_region(source, region, encoding)


def format_to_encoding(format: Optional[str]) -> str:
    """Translate format extension (e.g., `tiff`) to encoding (e.g., `TILED`)."""
    if not format:
        return 'PNG'
    if format.lower() not in ['tif', 'tiff', 'png', 'jpeg', 'jpg']:
        raise ValidationError(f'Format {format!r} is not valid. Try `png`, `jpeg`, or `tif`')
    if format.lower() in ['tif', 'tiff']:
        return 'TILED'
    return format.upper()  # jpeg, png


def get_frames(source: FileTileSource):
    """Return lists of channels/bands per frame index.

    Example Data
    ------------

    { frames: [
        { frame: 'Frame 1', bands: [
            {'index': 1, 'frame': 0, 'name': 'red'},
            {'index': 2, 'frame': 0, 'name': 'green'},
            {'index': 3, 'frame': 0, 'name': 'blue'},
        ]}
    ]}

    { frames: [
        { frame: 'Frame 1', bands: [{'index': 1, 'frame': 0, 'name': 'NUCLEI'}, ...]},
        { frame: 'Frame 2', bands: [{'index': 1, 'frame': 1, 'name': 'CD4'}, ...] },
        ...
    ]}


    """
    frame_data = source.getMetadata().get('frames', [])
    if not frame_data:
        # Single frame image
        bands = source.getBandInformation()
        frame = {
            'frame': 'Frame null',
            'bands': [
                {'index': k, 'frame': None, 'name': v.get('interpretation', '')}
                for k, v in bands.items()
            ],
        }
        frames = [frame]
    else:
        frames = {}
        for channel in frame_data:
            fid = channel['Frame']
            frames.setdefault(fid, [])
            frames[fid].append(
                {
                    'index': channel['Index'],
                    'frame': fid,
                    'name': channel.get('Name', ''),
                }
            )
        frames = [{'frame': f'Frame {i}', 'bands': v} for i, v in frames.items()]
    return {'frames': frames}



from drf_yasg import openapi

projection = openapi.Parameter(
    'projection',
    openapi.IN_QUERY,
    description='The projection in which to open the image (try `EPSG:3857`).',
    type=openapi.TYPE_STRING,
)
source = openapi.Parameter(
    'source',
    openapi.IN_QUERY,
    description='The source to use when opening the image. Use the `large-image/sources` endpoint to list the available sources.',
    type=openapi.TYPE_STRING,
)

BASE = [projection, source]

z = openapi.Parameter(
    'z',
    openapi.IN_PATH,
    description='The Z level of the tile. May range from [0, levels], where 0 is the lowest resolution, single tile for the whole source.',
    type=openapi.TYPE_INTEGER,
)
x = openapi.Parameter(
    'x',
    openapi.IN_PATH,
    description='The 0-based X position of the tile on the specified Z level.',
    type=openapi.TYPE_INTEGER,
)
y = openapi.Parameter(
    'y',
    openapi.IN_PATH,
    description='The 0-based Y position of the tile on the specified Z level.',
    type=openapi.TYPE_INTEGER,
)

# Style Parameters
palette = openapi.Parameter(
    'palette',
    openapi.IN_QUERY,
    description='The color palette to map the band values (named Matplotlib colormaps or palettable palettes). `cmap` alias supported.',
    type=openapi.TYPE_STRING,
)
band = openapi.Parameter(
    'band',
    openapi.IN_QUERY,
    description='The band number to use.',
    type=openapi.TYPE_INTEGER,
)
vmin = openapi.Parameter(
    'min',
    openapi.IN_QUERY,
    description='The minimum value for the color mapping.',
    type=openapi.TYPE_NUMBER,
)
vmax = openapi.Parameter(
    'max',
    openapi.IN_QUERY,
    description='The maximum value for the color mapping.',
    type=openapi.TYPE_NUMBER,
)
nodata = openapi.Parameter(
    'nodata',
    openapi.IN_QUERY,
    description='The value to map as no data (often made transparent).',
    type=openapi.TYPE_NUMBER,
)
scheme = openapi.Parameter(
    'scheme',
    openapi.IN_QUERY,
    description='This is either ``linear`` (the default) or ``discrete``. If a palette is specified, ``linear`` uses a piecewise linear interpolation, and ``discrete`` uses exact colors from the palette with the range of the data mapped into the specified number of colors (e.g., a palette with two colors will split exactly halfway between the min and max values).',
    type=openapi.TYPE_STRING,
)
style = openapi.Parameter(
    'style',
    openapi.IN_QUERY,
    description='Encoded string of JSON style following https://girder.github.io/large_image/tilesource_options.html#style',
    type=openapi.TYPE_STRING,
)

STYLE = [palette, band, vmin, vmax, nodata, scheme, style]

# Region Parameters
left = openapi.Parameter(
    'left', openapi.IN_QUERY, description='left', type=openapi.TYPE_NUMBER, required=True
)
right = openapi.Parameter(
    'right', openapi.IN_QUERY, description='right', type=openapi.TYPE_NUMBER, required=True
)
top = openapi.Parameter(
    'top', openapi.IN_QUERY, description='top', type=openapi.TYPE_NUMBER, required=True
)
bottom = openapi.Parameter(
    'bottom', openapi.IN_QUERY, description='bottom', type=openapi.TYPE_NUMBER, required=True
)
units = openapi.Parameter(
    'units',
    openapi.IN_QUERY,
    description='The projection/units of the region coordinates.',
    type=openapi.TYPE_STRING,
)

REGION = [left, right, top, bottom, units]

# Histogram Parameters
only = openapi.Parameter(
    'onlyMinMax',
    openapi.IN_QUERY,
    type=openapi.TYPE_BOOLEAN,
    default=False,
)
bins = openapi.Parameter(
    'bins',
    openapi.IN_QUERY,
    type=openapi.TYPE_INTEGER,
    default=256,
)
density = openapi.Parameter(
    'density',
    openapi.IN_QUERY,
    type=openapi.TYPE_BOOLEAN,
    default=False,
)
format = openapi.Parameter(
    'format',
    openapi.IN_QUERY,
    type=openapi.TYPE_STRING,
)

HISTOGRAM = [only, bins, density, format]


# Thumbnail Parameters
max_width = openapi.Parameter(
    'max_width',
    openapi.IN_QUERY,
    description='maximum width in pixels.',
    type=openapi.TYPE_INTEGER,
    default=256,
)
max_height = openapi.Parameter(
    'max_height',
    openapi.IN_QUERY,
    description='maximum height in pixels.',
    type=openapi.TYPE_INTEGER,
    default=256,
)

THUMBNAIL = [max_height, max_width]



from rest_framework import serializers


class TileMetadataSerializer(serializers.Serializer):

    levels = serializers.IntegerField(
        help_text='Number of zoom levels in the image.',
        min_value=1,
        read_only=True,
    )
    size_x = serializers.IntegerField(
        help_text='Image size in the X direction.',
        min_value=1,
        read_only=True,
        source='sizeX',
    )
    size_y = serializers.IntegerField(
        help_text='Image size in the Y direction.',
        min_value=1,
        read_only=True,
        source='sizeY',
    )
    tile_size = serializers.IntegerField(
        help_text='Size of the square tiles the image is composed of.',
        min_value=1,
        read_only=True,
        source='tileWidth',
    )
    geospatial = serializers.BooleanField(
        help_text='Whether the tile source is geospatial',
        read_only=True,
        source='dli_geospatial',
    )
    additional_metadata = serializers.JSONField(
        help_text='Any additional metadata on the tile source.',
        read_only=True,
        source='getMetadata',
    )



import json
import pathlib
from typing import Any, Optional, Union

from large_image.exceptions import TileSourceError
from large_image.tilesource import FileTileSource
from rest_framework.exceptions import APIException, ValidationError
from rest_framework.request import Request

from django_large_image import tilesource, utilities

CACHE_TIMEOUT = 60 * 60 * 2


class LargeImageMixinBase:
    def get_path(self, request: Request, pk: int = None) -> Union[str, pathlib.Path]:
        """Return path on disk to image file (or VSI str).

        This can be overridden downstream to implement custom FUSE, etc.,
        interfaces.

        Returns
        -------
        str : The local file path to pass to large_image

        """
        raise NotImplementedError(
            'You must implement `get_path` on your viewset.'
        )  # pragma: no cover

    def get_query_param(self, request: Request, key: str, default: Optional[Any] = '') -> str:
        return request.query_params.get(key, str(default))

    def get_style(self, request: Request) -> dict:
        # Check for url encoded style JSON
        if 'style' in request.query_params:
            try:
                style = json.loads(self.get_query_param(request, 'style'))
            except json.JSONDecodeError as e:
                raise ValidationError(
                    f'`style` query parameter is malformed and likely not properly URL encoded: {e}'
                )
        # else, fallback to supported query parameters for viewing a sinlge band
        else:
            band = int(self.get_query_param(request, 'band', 0))
            style = None
            if band:  # bands are 1-indexed
                style = {'band': band}
                bmin = self.get_query_param(request, 'min')
                bmax = self.get_query_param(request, 'max')
                if not utilities.param_nully(bmin):
                    style['min'] = bmin
                if not utilities.param_nully(bmax):
                    style['max'] = bmax
                palette = self.get_query_param(
                    request, 'palette', self.get_query_param(request, 'cmap')
                )
                if not utilities.param_nully(palette):
                    style['palette'] = palette
                nodata = self.get_query_param(request, 'nodata')
                if not utilities.param_nully(nodata):
                    style['nodata'] = nodata
                scheme = self.get_query_param(request, 'scheme')
                if not utilities.param_nully(scheme):
                    style['scheme'] = scheme
        return style

    def open_tile_source(
        self, request: Request, path: Union[str, pathlib.Path], **kwargs
    ) -> FileTileSource:
        """Override to open with a specific large-image source."""
        source = self.get_query_param(request, 'source')
        # error: "get_tilesource_from_path" gets multiple values for keyword
        # argument "source" Found 1 error in 1 file (checked 15 source files)
        return tilesource.get_tilesource_from_path(path, source=source, **kwargs)  # type: ignore

    def open_image(
        self,
        request: Request,
        path: Union[str, pathlib.Path],
        encoding: str = None,
        style: Union[bool, dict, str] = True,
    ) -> FileTileSource:
        projection = self.get_query_param(request, 'projection')
        kwargs = {}
        if encoding:
            kwargs['encoding'] = encoding
        if encoding:
            kwargs['encoding'] = encoding
        if style:
            if isinstance(style, bool) and style:
                _style = json.dumps(self.get_style(request))
            elif isinstance(style, dict) and style:
                _style = json.dumps(style)
            if isinstance(_style, str) and not utilities.param_nully(_style):
                kwargs['style'] = _style
        if projection:
            kwargs['projection'] = projection
        return self.open_tile_source(request, path, **kwargs)

    def get_tile_source(
        self,
        request: Request,
        pk: int = None,
        encoding: str = None,
        style: Union[bool, dict] = True,
    ) -> FileTileSource:
        """Return the built tile source."""
        try:
            return self.open_image(
                request, self.get_path(request, pk), encoding=encoding, style=style
            )
        except TileSourceError as e:
            # Raise 500 server error if tile source failed to open
            raise APIException(str(e))



from django.http import HttpResponse
from django.utils.decorators import method_decorator
from django.views.decorators.cache import cache_page
from drf_yasg.utils import swagger_auto_schema
from rest_framework.decorators import action
from rest_framework.exceptions import ValidationError
from rest_framework.request import Request
from rest_framework.response import Response

from django_large_image import tilesource, utilities
from django_large_image.rest import params
from django_large_image.rest.base import CACHE_TIMEOUT, LargeImageMixinBase

thumbnail_summary = 'Returns thumbnail of full image.'
thumbnail_parameters = params.BASE + params.THUMBNAIL + params.STYLE
region_summary = 'Returns region tile binary.'
region_parameters = params.BASE + params.REGION
pixel_summary = 'Returns single pixel.'
pixel_parameters = params.BASE + [params.left, params.top] + params.STYLE
histogram_summary = 'Returns histogram'
histogram_parameters = params.BASE + params.HISTOGRAM


class DataMixin(LargeImageMixinBase):
    @method_decorator(cache_page(CACHE_TIMEOUT))
    def thumbnail(self, request: Request, pk: int = None, format: str = None) -> HttpResponse:
        encoding = tilesource.format_to_encoding(format)
        width = int(self.get_query_param(request, 'max_width', 256))
        height = int(self.get_query_param(request, 'max_height', 256))
        source = self.get_tile_source(request, pk, encoding=encoding)
        thumb_data, mime_type = source.getThumbnail(encoding=encoding, width=width, height=height)
        return HttpResponse(thumb_data, content_type=mime_type)

    @swagger_auto_schema(
        method='GET',
        operation_summary=thumbnail_summary,
        manual_parameters=thumbnail_parameters,
    )
    @action(detail=False, url_path='thumbnail.png')
    def thumbnail_png(self, request: Request, pk: int = None) -> HttpResponse:
        return self.thumbnail(request, pk, format='png')

    @swagger_auto_schema(
        method='GET',
        operation_summary=thumbnail_summary,
        manual_parameters=thumbnail_parameters,
    )
    @action(detail=False, url_path='thumbnail.jpeg')
    def thumbnail_jpeg(self, request: Request, pk: int = None) -> HttpResponse:
        return self.thumbnail(request, pk, format='jpeg')

    def region(self, request: Request, pk: int = None, format: str = None) -> HttpResponse:
        """Return the region tile binary from world coordinates in given EPSG.

        Note
        ----
        Use the `units` query parameter to inidicate the projection of the given
        coordinates. This can be different than the `projection` parameter used
        to open the tile source. `units` defaults to `EPSG:4326` for geospatial
        images, otherwise, must use `pixels`.

        """
        source = self.get_tile_source(request, pk)
        units = self.get_query_param(request, 'units')
        encoding = tilesource.format_to_encoding(format)
        left = float(self.get_query_param(request, 'left'))
        right = float(self.get_query_param(request, 'right'))
        top = float(self.get_query_param(request, 'top'))
        bottom = float(self.get_query_param(request, 'bottom'))
        path, mime_type = tilesource.get_region(
            source,
            left,
            right,
            bottom,
            top,
            units,
            encoding,
        )
        if not path:
            raise ValidationError(
                'No output generated, check that the bounds of your ROI overlap source imagery and that your `projection` and `units` are correct.'
            )
        tile_binary = open(path, 'rb')
        return HttpResponse(tile_binary, content_type=mime_type)

    @swagger_auto_schema(
        method='GET',
        operation_summary=region_summary,
        manual_parameters=region_parameters,
    )
    @action(detail=False, url_path=r'region.tif')
    def region_tif(self, request: Request, pk: int = None) -> HttpResponse:
        return self.region(request, pk, format='tif')

    @swagger_auto_schema(
        method='GET',
        operation_summary=region_summary,
        manual_parameters=region_parameters + params.STYLE,
    )
    @action(detail=False, url_path=r'region.png')
    def region_png(self, request: Request, pk: int = None) -> HttpResponse:
        return self.region(request, pk, format='png')

    @swagger_auto_schema(
        method='GET',
        operation_summary=region_summary,
        manual_parameters=region_parameters + params.STYLE,
    )
    @action(detail=False, url_path=r'region.jpeg')
    def region_jpeg(self, request: Request, pk: int = None) -> HttpResponse:
        return self.region(request, pk, format='jpeg')

    @swagger_auto_schema(
        method='GET',
        operation_summary=pixel_summary,
        manual_parameters=pixel_parameters,
    )
    @action(detail=False)
    def pixel(self, request: Request, pk: int = None) -> Response:
        left = int(self.get_query_param(request, 'left'))
        top = int(self.get_query_param(request, 'top'))
        source = self.get_tile_source(request, pk)
        metadata = source.getPixel(region={'left': left, 'top': top, 'units': 'pixels'})
        return Response(metadata)

    @swagger_auto_schema(
        method='GET',
        operation_summary=histogram_summary,
        manual_parameters=histogram_parameters,
    )
    @action(detail=False)
    def histogram(self, request: Request, pk: int = None) -> Response:
        only_min_max = not utilities.param_nully(self.get_query_param(request, 'onlyMinMax', False))
        density = not utilities.param_nully(self.get_query_param(request, 'density', False))
        kwargs = dict(
            onlyMinMax=only_min_max,
            bins=int(self.get_query_param(request, 'bins', 256)),
            density=density,
            format=self.get_query_param(request, 'format'),
        )
        source = self.get_tile_source(request, pk, style=False)
        result = source.histogram(**kwargs)
        if 'histogram' in result:
            result = result['histogram']
            for entry in result:
                for key in {'bin_edges', 'hist', 'range'}:
                    if key in entry:
                        entry[key] = [float(val) for val in list(entry[key])]
                for key in {'min', 'max', 'samples'}:
                    if key in entry:
                        entry[key] = float(entry[key])
        return Response(result)


class DataDetailMixin(DataMixin):
    @swagger_auto_schema(
        method='GET',
        operation_summary=thumbnail_summary,
        manual_parameters=thumbnail_parameters,
    )
    @action(detail=True, url_path='thumbnail.png')
    def thumbnail_png(self, request: Request, pk: int = None) -> HttpResponse:
        return super().thumbnail_png(request, pk)

    @swagger_auto_schema(
        method='GET',
        operation_summary=thumbnail_summary,
        manual_parameters=thumbnail_parameters,
    )
    @action(detail=True, url_path='thumbnail.jpeg')
    def thumbnail_jpeg(self, request: Request, pk: int = None) -> HttpResponse:
        return super().thumbnail_jpeg(request, pk)

    @swagger_auto_schema(
        method='GET',
        operation_summary=region_summary,
        manual_parameters=region_parameters,
    )
    @action(detail=True, url_path=r'region.tif')
    def region_tif(self, request: Request, pk: int = None) -> HttpResponse:
        return super().region_tif(request, pk)

    @swagger_auto_schema(
        method='GET',
        operation_summary=region_summary,
        manual_parameters=region_parameters + params.STYLE,
    )
    @action(detail=True, url_path=r'region.png')
    def region_png(self, request: Request, pk: int = None) -> HttpResponse:
        return super().region_png(request, pk)

    @swagger_auto_schema(
        method='GET',
        operation_summary=region_summary,
        manual_parameters=region_parameters + params.STYLE,
    )
    @action(detail=True, url_path=r'region.jpeg')
    def region_jpeg(self, request: Request, pk: int = None) -> HttpResponse:
        return super().region_jpeg(request, pk)

    @swagger_auto_schema(
        method='GET',
        operation_summary=pixel_summary,
        manual_parameters=pixel_parameters,
    )
    @action(detail=True)
    def pixel(self, request: Request, pk: int = None) -> Response:
        return super().pixel(request, pk)

    @swagger_auto_schema(
        method='GET',
        operation_summary=histogram_summary,
        manual_parameters=histogram_parameters,
    )
    @action(detail=True)
    def histogram(self, request: Request, pk: int = None) -> Response:
        return super().histogram(request, pk)



# flake8: noqa: F401
from django_large_image.rest.standalone import ListColormapsView, ListTileSourcesView
from django_large_image.rest.viewsets import (
    LargeImageDetailMixin,
    LargeImageFileDetailMixin,
    LargeImageMixin,
    LargeImageVSIFileDetailMixin,
)



import logging

from django.apps import AppConfig
from django.conf import settings
import large_image

logger = logging.getLogger(__name__)


class DjangoLargeImageConfig(AppConfig):
    name = 'django_large_image'
    verbose_name = 'Django Large Image'
    default_auto_field = 'django.db.models.BigAutoField'

    def ready(self):
        # Set up memcached with large_image
        if hasattr(settings, 'MEMCACHED_URL') and settings.MEMCACHED_URL:
            large_image.config.setConfig('cache_memcached_url', settings.MEMCACHED_URL)
            if (
                hasattr(settings, 'MEMCACHED_USERNAME')
                and settings.MEMCACHED_USERNAME
                and hasattr(settings, 'MEMCACHED_PASSWORD')
                and settings.MEMCACHED_PASSWORD
            ):
                large_image.config.setConfig(
                    'cache_memcached_username', settings.MEMCACHED_USERNAME
                )
                large_image.config.setConfig(
                    'cache_memcached_password', settings.MEMCACHED_PASSWORD
                )
            large_image.config.setConfig('cache_backend', 'memcached')
            logger.info('large_image is configured for memcached.')



from io import open as io_open
from pathlib import Path

from setuptools import find_packages, setup

readme_file = Path(__file__).parent / 'README.md'
if readme_file.exists():
    with io_open(readme_file, 'r', encoding='utf-8') as f:
        long_description = f.read()
else:
    # When this is first installed in development Docker, README.md is not available
    long_description = ''

setup(
    name='django-large-image',
    version='0.6.0',
    description='Dynamic tile server in Django built on top of large-image (and GDAL)',
    long_description=long_description,
    long_description_content_type='text/markdown',
    license='Apache 2.0',
    author='Kitware, Inc.',
    author_email='kitware@kitware.com',
    url='https://github.com/girder/django-large-image',
    keywords='',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Framework :: Django :: 3.0',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python',
    ],
    python_requires='>=3.8',
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        'django',
        'djangorestframework',
        'drf-yasg',
        'filelock',
        'large-image>=1.14',
    ],
    extras_require={
        'colormaps': [
            'matplotlib',
            'cmocean',
        ],
    },
)



from pkg_resources import DistributionNotFound, get_distribution

try:
    __version__ = get_distribution('django-large-image').version
except DistributionNotFound:  # pragma: no cover
    # package is not installed
    __version__ = ''



from django.urls import path

from django_large_image import rest

urlpatterns = [
    path(
        'api/large-image/sources',
        rest.ListTileSourcesView.as_view(),
        name='large-image-sources',
    ),
    path(
        'api/large-image/colormaps',
        rest.ListColormapsView.as_view(),
        name='large-image-colormaps',
    ),
]



from functools import wraps
import pathlib
from typing import Union

from django.db.models.fields.files import FieldFile
from rest_framework.exceptions import APIException
from rest_framework.request import Request

from django_large_image import utilities
from django_large_image.rest.data import DataDetailMixin, DataMixin
from django_large_image.rest.metadata import MetaDataDetailMixin, MetaDataMixin
from django_large_image.rest.tiles import TilesDetailMixin, TilesMixin


class LargeImageMixin(DataMixin, MetaDataMixin, TilesMixin):
    """Mixin for large-image endpoints on a non-detail viewset.

    This is for use on a standalone ``ViewSet`` class with no detial views and
    not associated to a model. The ``detail`` attribute is set to ``False``
    for the django-rest-framework ``@action``.

    Subclasses must implement `get_path()`:

    .. code:: python

        def get_path(self, request: Request, pk: int = None):
            instance = Model.objects.get(pk=pk)
            return instance.file.name

    """

    pass


class LargeImageDetailMixin(DataDetailMixin, MetaDataDetailMixin, TilesDetailMixin):
    """Mixin class for large-image endpoints on detail viewset.

    Subclasses must implement `get_path()`:

    .. code:: python

        def get_path(self, request: Request, pk: int = None):
            instance = Model.objects.get(pk=pk)
            return instance.file.name

    """

    pass


class LargeImageFileDetailMixin(LargeImageDetailMixin):
    """Mixin specifically for detail viewsets that have a file field on the model.

    Define `FILE_FIELD_NAME` as the string name of the file field on the
    subclassed ViewSet's model.

    This mixin assumes `get_object()` is present - thus for ViewSets
    """

    FILE_FIELD_NAME: str = ''

    def get_field_file(self) -> FieldFile:
        """Get `FileField` using `FILE_FIELD_NAME`."""
        try:
            return getattr(self.get_object(), self.FILE_FIELD_NAME)  # type: ignore
        except (AttributeError, TypeError):
            # Raise 500 server error
            raise APIException('`FILE_FIELD_NAME` not properly set on viewset class.')

    @wraps(LargeImageDetailMixin.get_path)
    def get_path(self, request: Request, pk: int = None) -> Union[str, pathlib.Path]:
        return utilities.field_file_to_local_path(self.get_field_file())


class LargeImageVSIFileDetailMixin(LargeImageFileDetailMixin):
    USE_VSI: bool = True

    @wraps(LargeImageFileDetailMixin.get_path)
    def get_path(self, request: Request, pk: int = None) -> Union[str, pathlib.Path]:
        """Wrap get_path with VSI support."""
        field_file = self.get_field_file()
        if self.USE_VSI:
            with utilities.patch_internal_presign(field_file):
                # Grab URL and pass back VSI path
                # DO NOT return here to make sure this context is cleared
                vsi = utilities.make_vsi(field_file.url)
            return vsi
        # Checkout file locally if no VSI
        return LargeImageFileDetailMixin.get_path(self, request, pk)
